"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import configargparse
import os
import yaml


class C_Parser:

    @staticmethod
    def required_dir_exists(path):
        # Check input does not contain spaces
        if not os.path.exists(path):
            msg = " No buildable directory found at {}".format(path)
            raise configargparse.ArgumentTypeError(msg)
        return path

    @staticmethod
    def str2bool(v):
        # convert possible options for boolean values into a bool itself.
        if isinstance(v, bool):
            return v
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise configargparse.ArgumentTypeError('Boolean value expected.')

    def __init__(self):
        # from: https://realpython.com/command-line-interfaces-python-argparse/

        self.parser = configargparse.ArgParser(description='Capture all build files and send for analysis.', )

        self.parser.add_argument('-c', '--config', is_config_file=True, help='Configuration file path.')

        # Add the arguments
        self.parser.add_argument('-bc', '--build_cmd',
                                 required=True,
                                 type=str,
                                 help='Command used to execute the build')

        self.parser.add_argument('-d', '--build_dir',
                                 required=True,
                                 type=C_Parser.required_dir_exists,
                                 help='Directory from which to run build. Absolute path (not relative) information is required')

        self.parser.add_argument('-Cov', '--coverity_root',
                                 metavar='coverity_root',
                                 dest='coverity_root',
                                 default='',
                                 type=str,
                                 help='Base directory for coverity.If not specified, blackduck-c-cpp downloads latest mini coverity package from GCP for authorized Black Duck customers for Black Duck versions >= 2021.10. Absolute path (not relative) information is required')

        self.parser.add_argument('-Cd', '--cov_output_dir',
                                 metavar='cov_output_dir',
                                 dest='cov_output_dir',
                                 default='',
                                 type=str,
                                 help='Target directory for coverity output files.If not specified, defaults to user_home/.synopsys/blackduck-c-cpp/output/project_name. Absolute path (not relative) information is required')

        self.parser.add_argument('-od', '--output_dir',
                                 metavar='output_dir',
                                 dest='output_dir',
                                 default='',
                                 type=str,
                                 help='Target directory for blackduck-c-cpp output files.If not specified, defaults to user_home/.synopsys/blackduck-c-cpp/output/project_name. Absolute path (not relative) information is required')

        self.parser.add_argument('-s', '--skip_build',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='Skip build and use previously generated build data.')

        self.parser.add_argument('-v', '--verbose',
                                 metavar='verbose',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='verbose mode selection')

        self.parser.add_argument('-proj', '--project_name',
                                 required=True,
                                 dest='project_name',
                                 type=str,
                                 help='Black Duck project name')

        self.parser.add_argument('-vers', '--project_version',
                                 required=True,
                                 dest='project_version',
                                 type=str,
                                 help='Black Duck project version')

        self.parser.add_argument('-Cl', '--codelocation_name',
                                 dest='codelocation_name',
                                 type=str,
                                 help='Codelocation name')

        self.parser.add_argument('-bd', '--bd_url',
                                 required=True,
                                 metavar='bd_url',
                                 type=str,
                                 help='Black Duck URL')

        self.parser.add_argument('-a', '--api_token',
                                 dest='api_token',
                                 required=True,
                                 metavar='api_token',
                                 type=str,
                                 help='Black Duck API token')

        self.parser.add_argument('-as', '--additional_sig_scan_args',
                                 metavar='additional_sig_scan_args',
                                 dest='additional_sig_scan_args',
                                 type=str,
                                 default=None,
                                 help='Any additional args to pass to the signature scanner')

        self.parser.add_argument('-i', '--insecure',
                                 dest='insecure',
                                 metavar='insecure',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='Disable SSL verification so self-signed Black Duck certs will be trusted')

        self.parser.add_argument('-f', '--force',
                                 dest='force',
                                 metavar='force',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='In case of GCP failure, force use of older version of Coverity (if present)')

        self.parser.add_argument('-djs', '--disable_json_splitter',
                                 dest='disable_json_splitter',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='Disable the json splitter and always upload as a single scan')

        self.parser.add_argument('-si', '--scan_interval',
                                 dest='scan_interval',
                                 type=int,
                                 default=60,
                                 help='Set the number of seconds to wait between scan uploads in case of multiple scans')

        self.parser.add_argument('-jsl', '--json_splitter_limit',
                                 dest='json_splitter_limit',
                                 default=4750000000,
                                 metavar='json_splitter_limit',
                                 type=int,
                                 help='Set the limit for a scan size in bytes')

        self.parser.add_argument('-dg', '--debug',
                                 metavar='debug',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='Debug mode selection. Setting debug: True sends all the files we found to all matching types. By default, it will only send files not detected by package manager to BDBA and Signature matching.')

        self.parser.add_argument('-st', '--skip_transitives',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='Skipping all transitive dependencies')

        self.parser.add_argument('-sh', '--skip_includes',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='Skipping all .h & .hpp files from all types of scan')

        self.parser.add_argument('-sd', '--skip_dynamic',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='Skipping all .so files from all types of scan')

        self.parser.add_argument('-off', '--offline',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='store bdba and sig tar files and c_cpp_bdio2.jsonld to disk if offline mode is true')

        self.parser.add_argument('-md', '--modes',
                                 metavar='modes',
                                 dest='modes',
                                 type=str,
                                 default="ALL",
                                 help="comma separated list of modes to run - 'all'(default),'bdba','sig','pkg_mgr'")

        self.parser.add_argument('-uo', '--use_offline_files',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='use offline generated files for upload in online mode')

        self.parser.add_argument('-sc', '--scan_cli_dir',
                                 metavar='scan_cli_dir',
                                 dest='scan_cli_dir',
                                 default=None,
                                 type=str,
                                 help='scan cli directory. Absolute path (not relative) information is required')

        self.parser.add_argument('-Cc', '--cov_configure_args',
                                 metavar='cov_configure_args',
                                 dest='cov_configure_args',
                                 default={},
                                 type=yaml.safe_load,
                                 required=False,
                                 help='Additional configuration commands to cov-configure for different compilers. Inputs taken are of format {"compiler":"compiler-type"}. There is a way to use coverity template configuration to reduce number of template compiler configurations with wildcards: example: "--compiler *g++ --comptype gcc" for adding x86_64-pc-linux-gnu-g++')

        self.parser.add_argument('-ac', '--additional_coverity_params',
                                 metavar='additional_coverity_params',
                                 dest='additional_coverity_params',
                                 type=str,
                                 default=None,
                                 help='Any additional args to pass to coverity build command. example: "--record-with-source"')

        self.parser.add_argument('-es', '--expand_sig_files',
                                 type=C_Parser.str2bool,
                                 nargs='?',
                                 const=True,
                                 default=False,
                                 help='use expand_sig_files for creating exploded directory instead of tar in sig scanner mode')

        self.args = self.parser.parse_args()



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import os

from typing import Any, Dict, List, Union

PathLike = Union[os.PathLike, str]
# not a perfect JSON definition but no need for too fancy, it is mostly for documentation
# see https://stackoverflow.com/a/51294709
JsonDict = Dict[str, Any]

# TODO: Do a proper cleanup for custom types
#  Either start using TypedDicts after Alpine releases Python 3.8 or start using proper classes

# file paths following nested archives with left one being the toplevel archive/file path
FullPath = List[str]
# same as above but also states the archive/file type for each path
ExtendedFullPath = List[Dict[str, str]]

WhiteoutEntry = JsonDict
# WhiteoutEntry example:
# {
#     'object': '.wh.libformw.so.6',
#     'fullpath': ['bash.tar',
#                  'b7ff2716a6b0bd2caa3a4f4a84f8016913676782d5eb77526ad81be0a62b9619/layer.tar',
#                  'usr/lib/.wh.libformw.so.6'],
#     'extended-fullpath': [{
#         'type': 'tar',
#         'path': 'bash.tar'
#     }, {
#         'type': 'tar',
#         'path': 'b7ff2716a6b0bd2caa3a4f4a84f8016913676782d5eb77526ad81be0a62b9619/layer.tar'
#     }, {
#         'type': 'docker-whiteout',
#         'path': 'usr/lib/.wh.libformw.so.6'
#     }],
#     'timestamp': 1556193489,
#     'size': 0
# }

# TODO: add type hint for distro package match object

ExtendedObject = JsonDict
# ExtendedObject example:
# {
#     'name': 'mkmntdirs',
#     'fullpath': ['bash.tar',
#                  '0d2eaf4aa1a6112f04f5377ab5faee43703e2ded380b6109478f2db1fe0666bf/layer.tar', 'sbin/mkmntdirs'],
#     'extended-fullpath': [{
#                               'type': 'tar',
#                               'path': 'bash.tar'
#                           }, {
#                               'type': 'docker-layer',
#                               'path': '0d2eaf4aa1a6112f04f5377ab5faee43703e2ded380b6109478f2db1fe0666bf/layer.tar'
#                           }, {
#                               'type': 'elf',
#                               'path': 'sbin/mkmntdirs'
#                           }],
#     'timestamp': 1548276389,
#     'size': 13968,
#     'sha1': '2ad5d792e81a8feaf06a7ddc0dc8118e1e965250',
#     'confidence': 0.8,
#     'matching-method': 'distro-package-manager',
#     'binary-type': 'elf-executable-x86_64',
#     'package-name': 'alpine-baselayout',
#     'package-type': 'apk', <optional>
#     'package-version': '3.1.0-r3', <optional>
#     'distro': 'alpine', <optional>
#     'distro-version': '3.9.3', <optional>
#     'package-architecture': 'x86_64', <optional>
#     'type': 'distro-package',
#     'docker': {  <optional>
#         'image-id': '7f80652c1f4116a71fe90885c37c78feebd451fb4f50d7d323a0da209076c73a',
#         'layer-id': '0d2eaf4aa1a6112f04f5377ab5faee43703e2ded380b6109478f2db1fe0666bf/layer.tar'
#     }
# }

FindlibResult = JsonDict
# FindlibResult example:
# {
#     'lib': 'zsh',
#     'objects': ['zsh'],
#     'version': '5.6.2-r0',
#     'license': 'MIT',
#     'homepage': 'http://www.zsh.org',
#     'extended-objects':
#         [{
#              'name': 'zsh',
#              'fullpath': ['bash.tar', '3c3d21475/layer.tar',
#                           'bin/zsh'],
#              'extended-fullpath': [{
#                                        'type': 'tar',
#                                        'path': 'bash.tar'
#                                    }, {
#                                        'type': 'tar',
#                                        'path': '3c3d21475/layer.tar'
#                                    }, {
#                                        'type': 'elf',
#                                        'path': 'bin/zsh'
#                                    }],
#              'timestamp': 1545308087,
#              'size': 644536,
#              'exe-flags': ['no-fortify-source'],
#              'sha1': '8b7a1ecd3c27a05adad32366b382db64bc3a15ae',
#              'confidence': 0.9260504201680673,
#              'matching-method': 'signature',
#              'binary-type': 'elf-executable-x86_64',
#              'type': 'native'
#          }]
# }

FingerprintEntry = JsonDict
# FingerprintEntry example (not very comprehensive since there are PLENTY of optional fields):
# {
#     'added': 1384337904,
#     'aliases': ['http_server', 'apache_http_server', 'apache2'],
#     'homepage': 'https://httpd.apache.org/',
#     'lib': 'apache',
#     'license': 'Apache',
#     'proper-nvd': True,
#     'rodata': {
#         "string",
#         "another string",
#     },
#     'sources': ['/usr/sbin/apache2'],
#     'tags': ['protocol', 'server', 'http'],
#     'updated': 1384337904,
#     'vendor': 'apache',
#     'ver-regexp': 'Apache/(\\d+\\.\\d+\\.\\d+)'
# }

FingerprintDB = List[FingerprintEntry]



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import logging
import os
import tarfile
import re
import platform
import json
import glob
import shutil
import tempfile
import zipfile
import requests
import tqdm
import time
from blackduck_c_cpp.util import util
from blackduck_c_cpp.util.zip_file import MyZipFile
from blackduck_c_cpp.util import global_settings


class SigScanner:

    def __init__(self, cov_base_path, hub_api, files_notdetect_from_pkgmgr, project_name, project_version_name,
                 code_location_name, cov_emit_output_sig, blackduck_output_dir, offline_mode, scan_cli_dir,
                 use_offline_files, os_dist, additional_sig_scan_args, expand_sig_files, scan_interval,
                 json_splitter_limit,
                 disable_json_splitter, skip_includes):
        self.hub_api = hub_api
        self.sleep_interval = scan_interval  # time to wait between scans
        self.project_name = project_name
        self.project_version_name = project_version_name
        self.code_location_name = code_location_name + '-SIG'
        self.insecure = hub_api.insecure
        self.additional_sig_scan_args = additional_sig_scan_args
        self.expand_sig_files = expand_sig_files
        if self.additional_sig_scan_args:
            self.expand_sig_files = True if global_settings.snippet_match in self.additional_sig_scan_args.lower() else self.expand_sig_files
        self.json_splitter_limit = json_splitter_limit
        self.disable_json_splitter = disable_json_splitter
        self.skip_includes = skip_includes
        self.cov_emit_output_sig = cov_emit_output_sig
        self.offline_mode = offline_mode
        self.use_offline_files = use_offline_files
        self.os_dist = os_dist
        self.api_token = hub_api.api_token
        self.cov_base_path = cov_base_path
        self.blackduck_output_dir = blackduck_output_dir
        self.sig_scan_output_directory = os.path.join(blackduck_output_dir, 'sig_scan')
        self.sig_files_output_path = os.path.join(self.sig_scan_output_directory, 'sig_files')
        self.sig_scan_tar_output_directory = os.path.join(self.sig_scan_output_directory, 'tar_file')
        self.tar_output_path = os.path.join(self.sig_scan_tar_output_directory, 'sig_scan.tar.gz')
        self.scan_cli_home = scan_cli_dir
        self.files_notdetect_from_pkgmgr = files_notdetect_from_pkgmgr
        self.files_to_tar = None
        self.scan_cli_directory = None
        self.scan_cli_path = None
        self.cov_header_files = {}
        self.sig_scanner_found = False
        self.json_file = self.bdio_file = None
        self.upload_dry_run = self.use_offline_files
        self.proxy_settings = ''
        self.url_split = "://"
        self.op_sys = platform.system()

    def get_files_to_tar(self):
        """ collect all files to send to signature scanner"""
        if self.skip_includes:
            self.cov_emit_output_sig = [emit_file for emit_file in self.cov_emit_output_sig if not (
                    re.match(global_settings.hpp_pattern, emit_file.strip()) or re.match(global_settings.h_pattern,
                                                                                         emit_file.strip()))]
        self.files_to_tar = set(self.cov_emit_output_sig)
        if self.files_notdetect_from_pkgmgr:
            [[[self.files_to_tar.add(path) for path in paths if os.path.exists(path)] for paths in type.values()] for
             type in
             self.files_notdetect_from_pkgmgr.values()]

    def copy_files_to_dir(self):
        """
        write files to scan to exploded directory
        """
        self.get_files_to_tar()
        if not os.path.exists(self.sig_scan_output_directory):
            os.makedirs(self.sig_scan_output_directory)

        logging.info("The signature files to scan will be written to {}".format(self.sig_files_output_path))
        logging.info("Collecting files returned by cov-emit...")
        logging.info("Number of files in sig_files directory are {}".format(len(self.files_to_tar)))

        if os.path.exists(self.sig_files_output_path):
            shutil.rmtree(self.sig_files_output_path)
        files_count = 0
        for each_file in tqdm.tqdm(self.files_to_tar, total=len(self.files_to_tar)):
            dest_filepath = ''
            try:
                # exclude coverity configuration header files that are in the emit directory
                if os.path.join(self.blackduck_output_dir, 'emit') in each_file:
                    continue
                # split drive -> drive is mount point and path is in tail
                file_spath = os.path.splitdrive(each_file)[1]
                dest_filepath = self.sig_files_output_path + file_spath
                dest_folder = os.path.dirname(dest_filepath)

                if not os.path.exists(dest_folder):
                    os.makedirs(dest_folder)
                shutil.copy(each_file, dest_filepath)
                files_count += 1
            except IsADirectoryError:
                continue
            except PermissionError:
                continue
            except Exception as e:
                if len(dest_filepath) > 260 and self.op_sys == 'Windows':
                    logging.warning("You are using Windows and the file path exceeds the default limit of 260 characters. Check the character limit on your system and consider removing the limit to include this file in the scan. The rest of the scan will proceed.")
                    logging.warning("Not copying file to location: {}".format(dest_filepath))
                else:
                    logging.error("Exception is {}".format(e))
                continue
        logging.debug("Number of files added in sig_files directory are {}".format(files_count))

    def tar_files(self):
        """
        compress the files from run_cov_emit and from files_notdetect_from_pkgmgr into a tar archive
        os.path.exists() check done in emit_wrapper.py
        """
        self.get_files_to_tar()
        if not os.path.exists(self.sig_scan_tar_output_directory):
            os.makedirs(self.sig_scan_tar_output_directory)
        logging.info("The signature tar file to scan will be written to {}".format(self.tar_output_path))
        logging.info("Tarring files returned by cov-emit...")
        logging.info("Number of files in sig tar file is {}".format(len(self.files_to_tar)))
        with tarfile.open(self.tar_output_path, 'w:gz') as tar:
            for f in tqdm.tqdm(self.files_to_tar, total=len(self.files_to_tar)):
                tar.add(f)

    def get_sig_scanner_download_url(self):
        """
        construct the URL from which to download the signature scanner from Black Duck
        """
        sig_scanner_url = self.hub_api.hub.config['baseurl'] + '/download/scan.cli'
        if self.op_sys == 'Windows':
            sig_scanner_url += '-windows'
        if self.op_sys == 'Darwin':
            sig_scanner_url += '-macosx'
        sig_scanner_url += '.zip'
        return sig_scanner_url

    def find_any_sig_scanner(self):
        """
        check if the sig scanner for any black duck version is already on the system
        """
        try:
            self.scan_cli_directory = os.path.basename(
                sorted(glob.glob(os.path.join(self.scan_cli_home, 'scan.cli-*')), reverse=True)[0])
            self.scan_cli_path = os.path.join(self.scan_cli_home, self.scan_cli_directory)
            self.sig_scanner_found = True
        except IndexError:
            logging.warning("Scan cli not found at location: {}".format(self.scan_cli_home))

    def find_current_sig_scanner(self):
        """
        check if the sig scanner for the current black duck version is already on the system
        """
        self.scan_cli_directory = 'scan.cli-{}'.format(self.hub_api.hub.version_info['version'])
        self.scan_cli_path = os.path.join(self.scan_cli_home, self.scan_cli_directory)
        if os.path.exists(self.scan_cli_path):
            logging.info("Scan cli has been found, will not be downloaded. Location: {}".format(self.scan_cli_path))
            self.sig_scanner_found = True
            if self.offline_mode:
                logging.warning("Scan cli not found in offline mode at location {}".format(self.scan_cli_path))

    def download_sig_scanner(self):
        """
        Download the signature scanner
        """
        logging.debug("Downloading the signature scanner...")
        logging.debug("scan cli directory is {}".format(self.scan_cli_directory))
        logging.debug("scan cli home is {}".format(self.scan_cli_home))

        sig_scanner_url = self.get_sig_scanner_download_url()

        logging.info("The signature scanner will be downloaded to {}".format(self.scan_cli_home))
        headers = self.hub_api.hub.get_headers()
        # use requests bc blackduck.execute_get() doesn't allow redirects
        response = requests.get(sig_scanner_url, allow_redirects=True, headers=headers,
                                verify=not self.hub_api.hub.config['insecure'])

        open(os.path.join(self.scan_cli_home, 'sig-scanner.zip'), 'wb').write(response.content)
        with MyZipFile(os.path.join(self.scan_cli_home, 'sig-scanner.zip'), 'r') as zip_obj:
            zip_obj.extractall(self.scan_cli_home)

        self.sig_scanner_found = True

    @staticmethod
    def json_splitter(scan_path, max_node_entries=200000, max_scan_size=4750000000):
        """
        Splits a json file into multiple jsons so large scans can be broken up with multi-part uploads
        Modified from source: https://github.com/blackducksoftware/json-splitter

        :param scan_path: the path to the json file to split
        :param max_node_entries: the max node entries per scan
        :param max_scan_size: the max size of any single scan in bytes
        :return: a list of the newly generated json files to upload
        """
        new_scan_files = []
        file_split = "file://"

        with open(scan_path, 'r') as f:
            scan_data = json.load(f)

        data_length = len(scan_data['scanNodeList'])

        scan_size = sum(node['size'] for node in scan_data['scanNodeList'] if node['uri'].startswith(file_split))

        if scan_size < max_scan_size:
            return [scan_path]

        logging.info("Source directory greater than 4.75GB. Splitting the json.")

        scan_name = scan_data['name']
        scan_node_list = scan_data.pop('scanNodeList')
        scan_data.pop('scanProblemList')
        scan_data['scanProblemList'] = []
        base = scan_node_list[0]

        # Computing split points for the file
        scan_chunk_size = 0
        scan_chunk_nodes = 0
        split_at = [0]
        for i in range(0, data_length - 1):
            if scan_chunk_size + scan_node_list[i + 1][
                'size'] > max_scan_size or scan_chunk_nodes + 1 > max_node_entries:
                scan_chunk_size = 0
                scan_chunk_nodes = 0
                split_at.append(i)
            if scan_node_list[i]['uri'].startswith(file_split):
                scan_chunk_size = scan_chunk_size + scan_node_list[i]['size']
            scan_chunk_nodes += 1

        # Create array of split points shifting by one position
        split_to = split_at[1:]
        split_to.append(data_length - 2)  # don't include the last entry because it's the one for the entire tar archive

        # Splitting and writing the chunks
        new_scan_path = scan_path.split('.json')[0]
        for i in range(len(split_at)):
            logging.debug("Processing range {}, {}".format(split_at[i], split_to[i]))
            # for i in range(0, dataLength, maxNodeEntries):
            node_data = scan_node_list[split_at[i]:split_to[i]]
            if i > 0:
                node_data.insert(0, base)
            # scanData['baseDir'] = baseDir + "-" + str(i)
            scan_data['scanNodeList'] = node_data
            scan_data['name'] = scan_name + "-" + str(split_at[i])
            filename = new_scan_path + "-" + str(split_at[i]) + '.json'
            with open(filename, 'w') as outfile:
                json.dump(scan_data, outfile)
            scan_data.pop('scanNodeList')
            new_scan_files.append(filename)

        return new_scan_files

    def get_java_and_standalone_jar_paths(self):
        """
        Get the paths to the signature scanner java home and the sig scanner standalone jar
        """
        standalone_jar_path = None
        if self.os_dist.lower() == 'mac':
            java_path = os.path.join(self.scan_cli_path, 'jre', 'Contents', 'Home', 'bin', 'java')
        else:
            java_path = os.path.join(self.scan_cli_path, 'jre', 'bin', 'java')
        try:
            standalone_jar_path = glob.glob(os.path.join(self.scan_cli_path, 'lib', '*-standalone.jar'))[0]
            logging.debug("standalone jar path is {}".format(standalone_jar_path))
        except IndexError:
            logging.warning("scan cli jar not found at location: {}".format(self.scan_cli_path))
        return java_path, standalone_jar_path

    def build_sig_scanner_command(self):
        """
        Construct the signature scanner command to run
        """
        java_path, standalone_jar_path = self.get_java_and_standalone_jar_paths()
        sig_scanner_command = '"{}" -Done-jar.silent=true -Done-jar.jar.path="{}"{} -Xmx4096m -jar "{}" --no-prompt -v --project "{}" --release "{}" --name "{}" --individualFileMatching=ALL --binaryAllowedList h,hpp,cpp,c '.format(
            java_path,
            os.path.join(self.scan_cli_path, 'lib', 'cache', 'scan.cli.impl-standalone.jar'),
            self.proxy_settings,
            standalone_jar_path,
            self.project_name, self.project_version_name,
            self.code_location_name)

        if self.offline_mode:
            sig_scanner_command += '--dryRunWriteDir "{}" '.format(self.sig_scan_output_directory)
        else:
            if self.url_split in self.hub_api.hub.config['baseurl']:
                scheme, host = self.hub_api.hub.config['baseurl'].split(self.url_split)
                sig_scanner_command += '--scheme {} --host {} --port 443 '.format(scheme, host)
                if self.insecure:
                    sig_scanner_command += '--insecure '
                self.upload_dry_run = False
            else:
                logging.warning("Black Duck URL could not be parsed for scheme/host - will do dryrun scan instead")
                self.upload_dry_run = True

        if self.additional_sig_scan_args:
            sig_scanner_command += '{} '.format(self.parse_additional_args())

        output_dir = '"{}"'.format(
            self.sig_files_output_path if self.expand_sig_files else (self.sig_scan_tar_output_directory))
        sig_scanner_command += output_dir

        logging.info("sig scanner command is {}".format(sig_scanner_command))
        return sig_scanner_command

    def parse_additional_args(self):
        """
        Parse the self.additional_sig_scan_args string and return a string with the arguments, which may contain spaces,
        properly quoted.

        The entire string can't be quoted because it may contain multiple parameters. Only args should be quoted,
        not parameters.

        Example input: --parameter0 argument with spaces --parameter1 argumentwithoutspace --parameter2 argument--with--double--dashes --parameter3
        Example output: --parameter0  "argument with spaces" --parameter1  "argumentwithoutspace" --parameter2  "argument--with--double--dashes" --parameter3
        """
        additional_sig_scan_args_str_with_quotes = ''
        delim = '--'
        additional_sig_scan_args_lst = re.split('( {})'.format(delim), ' {}'.format(self.additional_sig_scan_args))

        while additional_sig_scan_args_lst:
            item = additional_sig_scan_args_lst.pop(0).strip()
            if item == delim:
                parameter_with_args = additional_sig_scan_args_lst.pop(0).split()
                parameter = delim + parameter_with_args.pop(0)
                additional_sig_scan_args_str_with_quotes += '{} '.format(
                    parameter)  # parameter does not need to be escaped, will never contain spaces

                # handle args, which may contain spaces that need to be escaped
                if parameter_with_args:
                    additional_sig_scan_args_str_with_quotes += ' "{}" '.format(' '.join(parameter_with_args))

        return additional_sig_scan_args_str_with_quotes

    def run_sig_scanner(self):
        """
        Scan the tar files created by tar_files()
        """
        logging.info("Running the signature scanner")
        if os.path.exists(os.path.join(self.sig_scan_output_directory, 'data')):
            shutil.rmtree(os.path.join(self.sig_scan_output_directory, 'data'))  # remove data directory if exists
        env = dict(os.environ)
        env['BD_HUB_TOKEN'] = self.api_token
        scan_cli_opts = os.getenv('SCAN_CLI_OPTS', '')
        if scan_cli_opts != '':
            self.proxy_settings = " {}".format(scan_cli_opts)
        sig_scanner_command = self.build_sig_scanner_command()
        util.run_cmd(sig_scanner_command, env=env, curdir=self.sig_scan_output_directory)
        logging.info("Finished running the signature scanner")

    def upload_sig_json_results(self):
        """
        Upload the json results of the sig scanner to Black Duck to create a BOM. Used if Black Duck version < 2021.10.0.
        """
        logging.info("Uploading signature scan file {}".format(self.json_file))
        try:
            sig_path = os.stat(self.sig_files_output_path) if self.expand_sig_files else os.stat(self.tar_output_path)
        except FileNotFoundError as e:
            util.error_and_exit(
                "FileNotFound error, please make sure yaml file is same in offline/online mode: {}".format(e))
        if self.disable_json_splitter or (sig_path.st_size < self.json_splitter_limit):
            response = self.hub_api.hub.upload_scan(self.json_file)
            if not response.ok:
                logging.error("Problem uploading the json file -- (Response({}): {})".format(response.status_code,
                                                                                             response.text))
        else:
            json_lst = self.json_splitter(self.json_file, max_scan_size=self.json_splitter_limit)
            for j in json_lst:
                response = self.hub_api.hub.upload_scan(j)
                logging.info('Uploading {} of {} json files'.format(str(json_lst.index(j) + 1), str(len(json_lst))))
                if json_lst.index(j) < len(json_lst) - 1:
                    time.sleep(self.sleep_interval)
                if not response.ok:
                    logging.error("Problem uploading the json file -- (Response({}): {})".format(response.status_code,
                                                                                                 response.text))

    def upload_sig_bdio_results(self):
        """
        Upload the bdio results of the sig scanner to Black Duck to create a BOM. Used if Black Duck version >= 2021.10.0.
        """

        ext = "intelligent-persistence-scans"
        endpoint = '{}/api/{}'.format(self.hub_api.hub.config['baseurl'], ext)
        logging.info("Uploading signature scan file {}".format(self.bdio_file))

        with tempfile.TemporaryDirectory() as tmpdirname:
            with zipfile.ZipFile(self.bdio_file, 'r') as zip_ref:
                zip_ref.extractall(tmpdirname)

            jsonld_headers_path = os.path.join(tmpdirname, 'bdio-header.jsonld')

            jsonld_files_lst = sorted(glob.glob('{}/bdio-entry*.jsonld'.format(tmpdirname)))

            # make sure headers file is uploaded first
            headers = self.hub_api.hub.get_headers()
            headers['Content-type'] = "application/vnd.blackducksoftware.intelligent-persistence-scan-1-ld-2+json"
            headers['User-Agent'] = "SCAN_CLI"
            logging.info("Uploading headers file with path {}".format(jsonld_headers_path))
            try:
                with open(jsonld_headers_path, 'rb') as f:
                    # use requests bc blackduck library tries to validate json data
                    response = requests.post(endpoint, headers=headers, data=f, verify=not self.insecure, timeout=600)
                if 'Location' not in response.headers:
                    logging.error("Location was not in response headers. Can not proceed.")
                    return None

                location = response.headers['Location']

                logging.info(
                    "Headers file uploaded successfully, proceeding to upload remaining scans using URL {}".format(
                        location))

                total_files_count = len(jsonld_files_lst)
                headers['X-BD-MODE'] = "append"
                headers['X-BD-DOCUMENT-COUNT'] = str(len(jsonld_files_lst))

                for f in jsonld_files_lst:
                    logging.info("Uploading {} of {} bdio entry jsonld files with name {}...".format(
                        str(jsonld_files_lst.index(f) + 1), str(total_files_count), f))
                    with open(f, 'rb') as d:
                        response = requests.put(location, headers=headers, data=d, verify=not self.insecure)
                        if not response.ok:
                            logging.error(
                                "Problem uploading the bdio file -- (Response({}): {})".format(response.status_code,
                                                                                               response.text))
                            self.logger.log_warning(response.headers)

                headers['X-BD-MODE'] = "finish"
                requests.put(location, headers=headers, verify=not self.insecure)

            except requests.exceptions.ReadTimeout as e:
                logging.error("Timeout error occurred when uploading signature scanner jsonld file {}".format(e))
                return None

    def find_dryrun_files(self):
        """
        Find the path to the dryrun file created by the sig scanner. Will be a BDIO if >=2021.10.0 or JSON if <2021.10.0.
        """
        hub_version = self.hub_api.get_hub_version()
        major, minor, patch = hub_version.split('.')

        if int(major) >= 2021 and int(minor) >= 10:
            bdio_files = glob.glob(os.path.join(self.sig_scan_output_directory, 'data', '*.bdio'))
            if bdio_files:
                bdio_files.sort(reverse=True)
                self.bdio_file = bdio_files[0]
                return True

        json_files = glob.glob(os.path.join(self.sig_scan_output_directory, 'data', '*.json'))
        if json_files:
            json_files.sort(reverse=True)
            self.json_file = json_files[0]
            return True
        return False

    def get_offline_files_in_dir(self):
        """
        For use if running with offline files. Find the exploded directory created in a previous run, and if it exists, check if
        the /data/ directory that contains the json or bdio scan file exists. If not, run the sig scanner.
        """
        # finding if tar file is present
        logging.info("Attempting to use offline files for signature scanner file at location: {}".format(
            self.sig_files_output_path))

        if os.path.exists(self.sig_files_output_path):
            logging.info("Found signature scanner files at location: {}".format(self.sig_files_output_path))
            return True
        else:
            logging.error(
                "Unable to find previously generated offline signature files for signature scanner, set use_offline_files to false to generate new ones.")
            return False

    def get_offline_tar_file(self):
        """
        For use if running with offline files. Find the tar file create in a previous run, and if it exists, check if
        the /data/ directory that contains the json or bdio scan file exists. If not, run the sig scanner.
        """
        # finding if tar file is present
        logging.info("Attempting to use offline files for signature scanner tar file at location: {}".format(
            self.tar_output_path))

        if os.path.exists(self.tar_output_path):
            logging.info("Found signature scanner tar file at location: {}".format(self.tar_output_path))
            return True
        else:
            logging.error(
                "Unable to find previously generated offline tar files for signature scanner, set use_offline_files to false to generate new ones.")
            return False

    def run(self):
        """
        entry point
        """
        # find the most recent available sig scanner if it is already downloaded
        if self.offline_mode:
            self.find_any_sig_scanner()
        if not self.offline_mode:
            self.find_current_sig_scanner()
            # download the sig scanner if necessary
            if not self.sig_scanner_found:
                self.hub_api.authenticate()
                self.download_sig_scanner()
            logging.info("Will wait {} seconds between scan uploads if there are multiple scans".format(
                str(self.sleep_interval)))
            logging.info("Results will be uploaded to {}".format(self.hub_api.hub.config['baseurl']))
            logging.info(
                "Results will be mapped to project {} version {} and code location {}".format(self.project_name,
                                                                                              self.project_version_name,
                                                                                              self.code_location_name))

        # tar the files or copy files to exploded directory in online mode
        if not self.use_offline_files:
            self.copy_files_to_dir() if self.expand_sig_files else self.tar_files()
        # if we don't find previously generated files in offline mode run sig scanner
        if self.use_offline_files:
            if self.expand_sig_files and not (self.get_offline_files_in_dir() or self.find_dryrun_files()):
                return
            if not self.expand_sig_files and not (self.get_offline_tar_file() or self.find_dryrun_files()):
                return

        # run the sig scanner
        if self.use_offline_files:
            # if previous files are generated in offline mode, upload them, else run signature scanner if scan-cli is present
            if self.find_dryrun_files():
                logging.info("Using old signature scan from previous offline scan at location {}".format(
                    self.sig_scan_output_directory))
            elif self.scan_cli_path:
                self.run_sig_scanner()
        elif self.scan_cli_path:
            self.run_sig_scanner()

        # upload the dryrun files
        if self.upload_dry_run:
            if not self.find_dryrun_files():
                logging.error(
                    "The signature scan file was not generated. There will not be signature matching results.")
            self.hub_api.authenticate()  # prevent timeout
            if self.bdio_file:
                self.upload_sig_bdio_results()
            elif self.json_file:
                self.upload_sig_json_results()
            else:
                logging.error("No dryrun scan file was found")

        if self.offline_mode:
            logging.debug("Signature scanner was run in offline mode. Files are stored at Location: {}".format(
                self.sig_scan_output_directory))



"""
Copyright (c) 2021 Synopsys, Inc.
Use subject to the terms and conditions of the Synopsys End User Software License and Maintenance Agreement.
All rights reserved worldwide.
"""

import os
import json
import logging
import re
from blackduck_c_cpp.util import global_settings


class CovJsonParser:
    """
    To get header, linker and executables from json generated from cov-emit-link tool
    """

    def __init__(self, json_file_path):
        self.json_file_path = json_file_path
        self.source_files_set = set()
        self.linker_files_set = set()
        self.executable_files_set = set()
        self.header_files_set = set()
        self.files_dict = {}
        self.file_id_kind_dict = {}

        if os.path.exists(json_file_path):
            try:
                with open(json_file_path, 'r') as f:
                    self.raw_bdio_json = json.load(f)
            except PermissionError:
                logging.error(
                    "you don't have permission to read file {}, please make sure you correct permissions".format(
                        self.json_file_path))
        else:
            self.raw_bdio_json = {}
            logging.error(
                "coverity json file not created. please verify if build completed successfully and json is written")

    def create_file_id_dict(self):
        """
        Function to create a dictionary with file_id as key and value as path
        {1: /g/b/c/a.so,2: /CMakeFiles/b/c/temp.a...}
        """
        try:
            files_list_dict = self.raw_bdio_json['files']
            for each_file in files_list_dict:
                file = each_file['case-preserved']
                file_id = each_file['id']
                if file_id in self.files_dict:
                    continue
                else:
                    self.files_dict[file_id] = file.strip()
        except KeyError:
            self.files_dict = {}

    def add_file_to_set(self, input_dict, file_id_type):
        """
        Function to disambiguate files to source, header and executables list
        param: file_id_type: can be 'file_id' or 'primary_file_id' from json
                input_dict: dictionary with file_id,kind as keys
                eg: {file:id: 1, kind: 'shared library'}
        """
        try:
            if len(self.file_id_kind_dict[input_dict[file_id_type]]) > 1:
                file = self.files_dict[input_dict[file_id_type]]
                if re.match(global_settings.so_pattern, file.strip()) or re.match(global_settings.dll_pattern,
                                                                                  file.strip()):
                    input_dict['kind'] = 'shared library'
                elif re.match(global_settings.a_pattern, file.strip()) or re.match(global_settings.lib_pattern,
                                                                                   file.strip()):
                    input_dict['kind'] = 'static library'
                elif re.match(global_settings.o_pattern, file.strip()) or re.match(global_settings.obj_pattern,
                                                                                   file.strip()):
                    input_dict['kind'] = 'object file'
                elif re.match(global_settings.c_pattern, file.strip()) or re.match(global_settings.cpp_pattern,
                                                                                   file.strip()) or \
                        re.match(global_settings.h_pattern, file.strip()) or re.match(global_settings.hpp_pattern,
                                                                                      file.strip()):
                    input_dict['kind'] = 'source file'

            if input_dict['kind'] == "source file":
                self.source_files_set.add(self.files_dict[input_dict[file_id_type]])
            elif input_dict['kind'] == "shared library" or input_dict['kind'] == "static library" or input_dict[
                'kind'] == "object file":
                self.linker_files_set.add(self.files_dict[input_dict[file_id_type]])
            elif input_dict['kind'] == "executable":
                self.executable_files_set.add(self.files_dict[input_dict[file_id_type]])
        except KeyError:
            pass

    def parse_for_files(self, unit_type):
        """
        Function to look at each dictionary in translation units and link units and
        call function to disambiguate files to source, header and executables list
        param: unit_type: 'translation units' or 'link-units'
        """
        try:
            tu_list_dict = self.raw_bdio_json[unit_type]
            for each_tu_dict in tu_list_dict:
                # logging.debug("each_tu_dict is {}".format(each_tu_dict))
                self.add_file_to_set(each_tu_dict, 'primary-file-id')
                for each_input_dict in each_tu_dict['input-files']:
                    # logging.debug("each_input_dict is {}".format(each_input_dict))
                    self.add_file_to_set(each_input_dict, 'file-id')
        except KeyError:
            logging.error("Unit not found in json: {}".format(unit_type))

    def add_file_to_dict(self, input_dict, file_id_type):
        """
        function to create dictionary with key as file-id amd value as list of all kinds
        params: file_id_type: can be 'file_id' or 'primary_file_id' from json
                input_dict: dictionary with file_id,kind as keys
        """
        if input_dict[file_id_type] in self.file_id_kind_dict:
            self.file_id_kind_dict[input_dict[file_id_type]].add(input_dict['kind'])
        else:
            self.file_id_kind_dict[input_dict[file_id_type]] = set([input_dict['kind']])

    def create_ambiguous_files_dict(self):
        """
        function to iterate through json and call another function to
        create dictionary with key as file-id amd value as list of all kinds
        for it from both translation and link units
        """
        units = ['translation-units', 'link-units']
        for each_unit in units:
            try:
                tu_list_dict = self.raw_bdio_json[each_unit]
                for each_tu_dict in tu_list_dict:
                    self.add_file_to_dict(each_tu_dict, 'primary-file-id')
                    for each_input_dict in each_tu_dict['input-files']:
                        self.add_file_to_dict(each_input_dict, 'file-id')
            except KeyError:
                logging.error("Unit not found in json: {}".format(each_unit))
                self.file_id_kind_dict = {}

    def get_header_files(self):
        """
        function to select header files from source files
        """
        return {file for file in self.source_files_set if file != ':' and (
                re.match(global_settings.h_pattern, file) or re.match(global_settings.hpp_pattern,
                                                                      file)) and os.path.exists(file)}

    def run(self):
        if self.raw_bdio_json:
            self.create_file_id_dict()
            self.create_ambiguous_files_dict()
            self.parse_for_files('translation-units')
            self.parse_for_files('link-units')
            self.header_files_set = self.get_header_files()
            logging.debug(
                "number of distinct executable files from coverity json are {}".format(len(self.executable_files_set)))
            logging.debug(
                "number of distinct linker files from coverity json are {}".format(len(self.linker_files_set)))
            logging.debug(
                "number of distinct header files from coverity json are {}".format(len(self.header_files_set)))
            return self.header_files_set, self.linker_files_set, self.executable_files_set
